'' +--------------------------------------------------------------------------+
'' | Cluso's PropBootSD                                                v0.xxx |
'' +--------------------------------------------------------------------------+
'' |  Authors:           Roger Williams / Michelli Scales (PropCMD)           |
'' |                     Jeff Ledger                      (PropDos)           |
'' |                     mctrivia                         (boot_loader eeprom)|
'' |                     Lonesock & Rokicki               (fsrw objects)      |
'' |                     Jeff Ledger & Mike Green         (fsrwFemto)         |
'' |                     Mike Green                       (Basic_I2C_Driver)  |
'' |                     Michael Park (mpark)             (Sphinx file xfer)  |
'' |  Apologies to any others I may have missed who have contributed.         |
'' |  Modifications:     (c) 2010 "Cluso99" (Ray Rodrick)                     |
'' |  License            MIT License - See end of file for terms of use       |
'' +--------------------------------------------------------------------------+
''
'' RR20100101   _rr055 modified to combine various functions from other sources
'' RR20100111   _rr058 add FTP (routine from Sphinx - acknowledgement to mpark)
''                     unfinished - kbd to send $00 !!
'' RR20100111   -rr059 use PC_Keyboard & PC_Text to use pins as passed

{{
PropCMD  Propeller Command Line SD card utility V1.0
by Roger Williams
Copyright (c) 2009 by Michelli Scales
See terms of use at end of this file

Based loosely on original PropDOS by Jeff Ledger

PropCMD implements a true parser which interprets arguments and
switches like /w regardless of the number of spaces separating
them, position on line, or order of switches.  PropCMD observes
the file system hidden attribute bit.

If there is a file on the SD named AUTOEXEC.BAT, when PropCMD
starts it will open the file, read the name of a file, and
load and run that file as if it were the argument of a SPIN
command.  Holding ESC down during boot inhibits this.  Note
that this is the opposite of PropDOS' autostart feature.
Also note this is not a true batch interpreter, and no other
commands can be included in autoexec.bat (so far).

PropCMD checks for AUTOEXEC.BAT before deciding to halt if
there is no keyboard, so it can be used to autoload a SD
program which does not in turn need a keyboard.

PropCMD COMMANDS:
-----------------

MOUNT
  Mount the SD.
  This is done automatically at startup if possible.

UNMOUNT
  Unmount the SD.
  The SD can then safely be changed.

DIR [*.ext] [/w] [/h] [/c] [|more] 
  Show the root directory of the SD.
  *.ext -- show only files with the indicated extension
  /w -- wide format
  /h -- show hidden files (based on attribute bit)
  /c -- show starting clusters
  |more -- pause between screens, ESC to quit
  ctrl-C cancels

TYPE filename [|more]
  Echo the file to the console. 
  |more -- pause between screens, ESC to quit
  ctrl-C cancels

DEL filename
  Delete the file.  Does not ask for confirmation.

COPY CON filename
  Echo lines of user input to the file until CTRL-Z pressed.
  ESC cancels the current input line without ending input.

SPIN filename
  Load the binary file and reboot the Propeller to run it.
  Forces .BIN extension whether it or any other is specified.

DUMP filename [|more]
  Display the file contents in hex and ASCII format.
  |more -- pause between screens, ESC to quit
  ctrl-C cancels

SD sectornum [|more]
  Sector Dump = display contents of the given SD sector
  in the Dump command format.
  
ROOTDIR
  Display the SD sector where the root directory begins

CLUST2SD clusternum
  Display the SD sector where DOS cluster begins

VER
  display the version

COLOR colornum
  Sets the display text color 0-7.
  
NUM number
  Display the hex and decimal equivalents of number.
  If number has no prefix or $, it is hexadecimal
  # prefix means decimal, e.g. B == $B == #11
  This format is used for all numeric arguments.

DIRA pin [value]
  If value is supplied, set the given pin DIRA to
  the value; otherwise display the pin DIRA value.

OUTA pin [value]
  If value is supplied, set the given pin OUTA to
  the value; otherwise display the pin OUTA value.

PULSE pin time
  Invert the given pin OUTA for the time in system
  waitcnt ticks.

INA pin
  Display the INA value for the pin.

FTP
  Do file transfer to/from PC using Sphinx method.

PropCMD does not attempt to SPIN commands it does not
recognize; it will reply Unknown Command.

Note that while this code observes the directory Hidden
Attribute bit, there are no functions to manipulate it.
fsrw does not expose this functionality and I would
rather work on a totally new filesystem capable of
opening multiple files than further patching to fsrw.

}}

CON
  ' I/O SETTINGS

{
'TriBlade
  _clkmode  = xtal1 + pll16x
'  _xinfreq  = 5_000_000
  _xinfreq  = 6_000_000
'  _xinfreq  = 6_500_000
'  _xinfreq  = 7_372_800
'  _clkmode  = xtal1 + pll8x
'  _xinfreq  = 13_500_000
'  _xinfreq  = 15_000_000
'SDcard
  spiDO         = 9             'SD pins
  spiClk        = 28
  spiDI         = 8
  spiCS         = 14
'Serial
  SI            = 31            'serial pins, mode & baudrate
  SO            = 30
  siomode       = 0
  baud          = 115200
'Keyboard
  kbdData       = 26 '-1            '26
  kbdClk        = 27 '-1            '27
'Video
  tvPins        = 12 '-1            '12,13,14,(15)
  screenrows    = 10            'for tv_text = 13, aigeneric = 25, pc_text = 10
}

'RamBlade
  _clkmode  = xtal1 + pll16x
  _xinfreq  = 6_500_000
'SDcard
  spiDO         = 24            'SD pins
  spiClk        = 26
  spiDI         = 25
  spiCS         = 19
'Serial
  SI            = 23            'serial pins, mode & baudrate
  SO            = 22
  siomode       = 0
  baud          = 115200
'Keyboard
  kbdData       = SI '-1            '26
  kbdClk        = baud '-1            '27
'Video
  tvPins        = SO '-1            '12,13,14,(15)
  screenrows    = 20            'for tv_text = 13, aigeneric = 25, pc_text = 10



{
   _clkmode  = xtal1 + pll16x
   _xinfreq  = 5_000_000
  ''
  '' Adjust for typical location of the keyboard.
  keyboard  = 26

  '' Adjust for the location of the video pingroup base
  video     = 12
'' Typical Hydra Configuration: 0-7 Left for NES/Audio Connections
  spiDO     = 16
  spiClk    = 17
  spiDI     = 18
  spiCS     = 19
'' Typical Demoboard/Protoboard Configuration
  spiDO     = 0
  spiClk    = 1
  spiDI     = 2
  spiCS     = 3
}

  bkspKey   = $C8
  escKey    = $CB
  ctlcKey   = $0263
  ctlzKey   = $027A
  
  USCORE    = $5F
  BKSP      = $08
  TAB       = $09
  LF        = $0A
  CR        = $0D
  CTLZ      = $1A
  ESC       = $1B
  SPC       = $20
  CTLC      = $03

  PAGESIZE = 32                                         'eeprom pagesize

OBJ
'    key   : "comboKeyboard"             ' Keyboard Driver
     key   : "PC_Keyboard_rr001"
     ''
     '' uncomment the desired video driver

'    text  : "tv_text"                   ' Parallax 13-line Video
     text  : "PC_Text_rr001"
     
{
     text  : "aigeneric_driver"          ' AiGeneric 25-line Video 
}
{    the aigeneric driver requires the files
        AiGeneric_Driver.spin
        AiGenericDriver_002.spin
        AiGenericDriver_TV.spin  ... and ...
        Font_ATARI.spin. }
     
'     fsrw  : "fsrwFemto_rr003a"           ' Modified SD routines
     fsrw  : "fsrwFemto_rr006"           ' Modified SD routines
     i2c   : "Basic_I2C_Driver"          'eeprom routines

VAR
     byte inputbuffer[64]
     byte inputfields
     byte inputparse[7]
     byte cmd
     byte arg1
     byte arg2
     byte arg3
     byte OptWide
     byte OptMore
     byte OptPRN
     byte OptWC
     byte Lcount
     byte EscLock
      
     byte Mounted
     
     byte dirbuf[fsrw#DIRSIZE] ' 
     byte fileext[4]
     byte numbuf[12]
     long ioControl[2]                                  '\\
     long i0Control_sector                              '// sdspiFemto passes ioControl[3] ==> sd sector no 
     byte secbuf[fsrw#SECTORSIZE] ' sector data buffer

'copy of variables in fsrwFemto
'-------------------------------
'   Variables concerning the open file.
   long x_fclust ' the current cluster number
   long x_filesize ' the total current size of the file
   long x_floc ' the seek position of the file
   long x_frem ' how many bytes remain in this cluster from this file
   long x_bufat ' where in the buffer our current character is
   long x_bufend ' the last valid character (read) or free position (write)
   long x_writelink ' the byte offset of the disk location to store a new cluster
   long x_direntry ' the byte address of the directory entry (if open for write)
   long x_fatptr ' the byte address of the most recently written fat entry
'   Variables used when mounting to describe the FAT layout of the card.
   long x_rootdir ' the byte address of the start of the root directory
   long x_rootdirend ' the byte immediately following the root directory.
   long x_dataregion ' the start of the data region, offset by two sectors
   long x_clustershift ' log base 2 of blocks per cluster
   long x_fat1 ' the block address of the fat1 space
   long x_totclusters ' how many clusters in the volume
   long x_sectorsperfat ' how many sectors per fat
'   Variables controlling the caching.
   long x_lastread ' the block address of the buf2 contents
   long x_dirty ' nonzero if buf2 is dirty
   long x_buf[2] ' only 2 longs

    byte  buff[512]    'SD buffer
    long  fbase        'SD first sector
    long  fsize        'SD filesize

    long wdcog
    long maincog
    long ss[20]     'stack

    byte buff2[1024]
    byte filename[16]           'pad to longs


PUB start | r, i, j, start_time

  key.start(kbdData, kbdClk)                            'start keyboard or PC (serial)
'  text.start(tvPins)                                    'start video    or PC (serial)
  text.start(tvPins, baud)                                    'start video    or PC (serial)

  waitcnt(clkfreq*1 + cnt)
  repeat until key.key == 0                             'ensure keyboard buffer empty

  crlf
  repeat 40
    text.out("-")
  crlf
  dosver

  r := fsrw.start(@ioControl)                           ' Start SPI driver (SD driver)
  Mounted := (\fsrw.mount(spiDO,spiClk,spiDI,spiCS) => 0)

  Prompt                                                'SD:> or No Volume:>
  text.str(string("Hit <esc> to bypass autoexec.bat,etc"))
  crlf

  waitcnt(clkfreq*3 + cnt)

  ' If keyboard present, check for the ESC key to bypass autostart
  ' Note the esclock, which prevents the escapes from scrolling
  ' down the screen if this bypass succeeds; once PropCMD starts
  ' escape is ignored until some other character is pressed.
  '
  esclock := -1
  if mounted
    r := true
'   if key.present
    if key.gotkey
      i:=key.key
      if i := ESC
        i := esckey
      if i == esckey
        PropCommand                                     '<esc> goto PropCommand (skip file stuff)
  else
    PropCommand                                         'no SD card - goto PropCommand

'------------------------------------------------------------------------------
'eeprom code

  r := FindFile(string("EEPROM_H.BIN"))                 'top eeprom code present?
  if r == 0
    text.str(string(" Programming Upper Eeprom..."))
    r := \ProgramEeprom($8000)                          'program upper eeprom from "EEPROM_H.BIN"
    if r == 0
      text.str(string(" done"))
      crlf
      if (FindFile(string("NODELETE.TXT")) <> 0)        'if nodelete.bin does not exist then
        fsrw.popen(string("EEPROM_H.BIN"),"d")          ' delete the eeprom file
        text.str(string(" file deleted"))
        crlf
    else
      text.str(string(" **failed**"))
      crlf

  r := FindFile(string("EEPROM_L.BIN"))                 'bottom eeprom code present?
  if r == 0
    text.str(string(" Programming Lower Eeprom..."))
    r := \ProgramEeprom($0000)                          'program lower eeprom from "EEPROM_L.BIN"
    if r == 0
      text.str(string(" done"))
      crlf
      if (FindFile(string("NODELETE.TXT")) <> 0)        'if nodelete.bin does not exist then
        fsrw.popen(string("EEPROM_L.BIN"),"d")          ' delete the eeprom file
        text.str(string(" file deleted"))
        crlf
    else
      text.str(string(" **failed**"))
      crlf

'------------------------------------------------------------------------------
'autoexec / binary boot code

  r := FindFile(string("AUTOEXEC.BAT"))                 'open and read "autoexec.bat"
  if r == 0
    r := fsrw.readSDCard(fbase, @buff, 128)             'get 128 bytes filename & params
    if r == 0
      text.str(string("  par: "))
      buff[12+1+32] := 0                                'terminate filename/parameters string
      text.str(@buff)
      buff[12] := 0                                     'terminate filename string 8.3
      crlf
      r := FindFile(@buff)                              'open binary boot file

  if r <> 0                                             'not found (autoexec.bat or invalid file parameter)
    r := FindFile(string("BOOTPROP.BIN"))               'open "bootprop.bin"
  if r == 0                                             'boot the binary file
    text.str(string(" Booting file..."))
    crlf
    waitcnt(clkfreq/10 + cnt)                           '100ms delay to clear text buffer
    r := fsrw.bootSDcard                                'boot the binary
    text.str(string("Boot error"))                      'should never get here!
    repeat                                              '<==================== so loop
  else
    PropCommand                                         'binary not found - goto PropCommand


Pri PropCommand
  ' MAIN COMMAND LINE INTERPRETER LOOP
  '
  repeat
    Prompt                                              'SD:> or No Volume:>
    if input(@inputbuffer,64) == cr
      ucase(@inputbuffer)
      parse(@inputbuffer)
      cmd := inputparse[0]
      arg1 := inputparse[1]
      arg2 := inputparse[2]
      lcount := 0
      if inputbuffer[0] <> 0
        \DosCmd
    else
      text.out(cr)

'------------------------------------------------------------------------------
{
PRI wd | i
  'wait for 10th of a sec
  waitcnt(cnt+clkfreq/2)

  'stop hanged main cog
  cogstop(maincog)

  'timeout so stop all cogs (this one last) - no point in loadTop as it failed and we may loop in an illegal state!
  repeat i from 0 to 7
    if i <> cogid
      cogstop (i)
  cogstop(cogid)                'finally stop this cog

PRI loadTop
  fsrw.bootEEPROM(fsrw#bootAddr + $8000)

Pri loadBottom
  fsrw.bootEEPROM(fsrw#bootAddr)
}

PRI FindFile (st) | r
  text.str(string(" Locating "))
  text.str(st)
  r := fsrw.popen(st,"r")                               'Open file for read (blank image)
  if r == 0
    fbase := fsrw.datablock                             'get the sectorno of the file
    fsize := fsrw.get_filesize                          'get size
    text.out(" ")
    text.dec(fbase)
    text.out(" ")
    text.dec(fsize)
  else
    text.str(string(" - not found"))
  crlf
  result := r


Pri ProgramEeprom (offset) | i, r, start_time
  repeat i from 0 to constant(($8000/PAGESIZE)-1)
    r:=fsrw.pread(@buff, PAGESIZE)
    if r == 0
      if i2c.WritePage(i2c#BootPin, i2c#EEPROM, offset+(i*PAGESIZE), @buff, PAGESIZE)
        abort                                           'error during write
      start_time := cnt                                 'prepare to check for a timeout
      repeat while i2c.WriteWait(i2c#BootPin, i2c#EEPROM, offset+(i*PAGESIZE))
        if cnt - start_time > clkfreq / 10
          abort                                         'waited more than 100ms

PRI CheckError(r)
  if r < 0
    text.str(string("***Error*** $"))
    text.hex(r,8)
    text.str(string(" ***HALTED***"))
    crlf
    fsrw.stop
    repeat                                              '<==============  LOOP on error

PRI crlf
  text.out(13)
  text.out(10)

Pri Prompt
  if Mounted
    text.str(string("SD:>"))
  else
    text.str(string("No Volume:>"))

PRI DosCmd |  r,i
  '
  ' This looks horrible, but when I implemented a lookup table and
  ' used a case statement it was neither smaller nor faster, and
  ' greatly complicated the task of adding a command.  
  '
  ' We start with commands that don't need a volume mounted
  '
  if strcomp(@inputbuffer+cmd,string("MOUNT"))
    if mounted
      text.str(string("Volume Already Mounted.",cr))
    else
      r := \fsrw.mount(spiDO,spiClk,spiDI,spiCS)
      if r => 0
        mounted := true
      else
        'hint: try plugging in the SD card
        text.str(string("Mount Failed: Error "))
        text.hex(r,9)
        text.out(cr)
    return
   
  if strComp(@inputbuffer+cmd,string("CLS"))
    text.out($00) 
    return
   
  if strComp(@inputbuffer+cmd,string("COLOR"))
    doscolor(numfromarg(@inputbuffer+arg1))
    return
   
  if strComp(@inputbuffer+cmd,string("REBOOT"))
    reboot

  if strComp(@inputbuffer+cmd,string("NUM"))
    i := numfromarg(@inputbuffer+arg1)
    text.out("$")
    text.hex(i,8)
    text.str(string(" = #"))
    text.dec(i)
    text.out(cr) 
    return

  if strComp(@inputbuffer+cmd,string("VER"))
    dosver
    return 

{{
  if strComp(@inputbuffer+cmd,string("DIRA"))
    if inputfields < 2
      abort
    i := numfromarg(@inputbuffer+arg1)
    if inputfields < 3
      'no value set, display
      text.str(string("Pin $"))
      text.hex(i,2)
      text.str(string(" DIRA = "))
      text.dec(dira[i])
      text.out(cr)
    else
      dira[i] := numfromarg(@inputbuffer+arg2)
    return            

  if strComp(@inputbuffer+cmd,string("OUTA"))
    if inputfields < 2
      abort
    i := numfromarg(@inputbuffer+arg1)
    if inputfields < 3
      'no value set, display
      text.str(string("Pin $"))
      text.hex(i,2)
      text.str(string(" OUTA = "))
      text.dec(outa[i])
      text.out(cr)
    else
      outa[i] := numfromarg(@inputbuffer+arg2)
    return            

  if strComp(@inputbuffer+cmd,string("PULSE"))
    if inputfields < 3
      abort
    i := numfromarg(@inputbuffer+arg1)
    r := numfromarg(@inputbuffer+arg2)
    !outa[i]
    waitcnt(r + cnt)
    !outa[i]
    return          

  if strComp(@inputbuffer+cmd,string("INA"))
    if inputfields < 2
      abort
    i := numfromarg(@inputbuffer+arg1)
    text.str(string("Pin $"))
    text.hex(i,2)
    text.str(string(" INA = "))
    text.dec(ina[i])
    text.out(cr)
    return
}}

  if mounted

    if strcomp(@inputbuffer+cmd,string("DIR"))
      dosdir
      return
     
    if strcomp(@inputbuffer+cmd,string("UNMOUNT"))
      \fsrw.unmount
      mounted := false 
      return
     
    if strComp(@inputbuffer+cmd,string("TYPE"))
      dostype
      return
     
    if strComp(@inputbuffer+cmd,string("DEL"))
      r := fsrw.popen(@inputbuffer+arg1,"d")
      if r < 0
        text.str(string("File "))
        text.str(@inputbuffer+arg1)
        text.str(string(" Not Found.",cr))
      else        
        text.str(string("File "))
        text.str(@inputbuffer+arg1)
        text.str(string(" Deleted.",CR)) 
      return
      
    if strComp(@inputbuffer+cmd,string("COPY"))
      if not strcomp(@inputbuffer+arg1,string("CON"))
        text.str(string("File copy not supported.",cr))
      else     
        r := fsrw.popen(@inputbuffer+arg2,"w")
        if r < 0
          text.str(string("Could not open file "))
          text.str(@inputbuffer+arg2)
          text.str(string(".",cr))
        repeat
          r := input(@inputbuffer,39)
          if r == ctlz
            fsrw.pclose
            text.out(cr)
            quit
          elseif r == esc
            text.str(string("/Line",cr))
          else
            i := 0
            repeat while inputbuffer[i] <> 0 
              fsrw.out(inputbuffer[i])
              i++
            fsrw.out(13) 
      return
             
    if strComp(@inputbuffer+cmd,string("DUMP"))
      dosdump
      return
     
    if strComp(@inputbuffer+cmd,string("DUMP2"))
      dosdump2
      return
     
    if strComp(@inputbuffer+cmd,string("X2"))
      dosX2
      return

    if strComp(@inputbuffer+cmd,string("XS"))
      dosXS
      return
     
    if strComp(@inputbuffer+cmd,string("ROOTDIR"))
      text.str(string("Root Directory at SD Sector $"))
      text.hex(fsrw.dos_rootdir,6)
      text.out(13)
      return
        
    if strComp(@inputbuffer+cmd,string("FAT"))
      text.str(string("FAT at SD Sector $"))
      text.hex(fsrw.dos_fat,6)
      text.out(13)
      return
     
    if strComp(@inputbuffer+cmd,string("SD"))
      secdump
      return
     
    if strComp(@inputbuffer,string("CLUST2SD"))
      clust2sd
      return
     
    if strComp(@inputbuffer,string("CHAIN"))
      doschain
      return
     
    if strComp(@inputbuffer,string("SPIN"))
      'attempt to autoexecute a .bin
      'copy filename base to dirbuf
      spin
      return
     
    if strComp(@inputbuffer,string("FTP"))
      'do file transfer to/from PC
      FileTransfer
      return

  text.str(string("Unknown or Invalid Command.",cr))
     
' individual command handlers

pri dosver
'  text.str(string("PropCommand (TriBlade) V0.060"))
  text.str(string("PropCommand (RamBlade) V0.060"))
  crlf

pri FileTransfer | cmd2, ch, i
  text.str(string("File Transfer",13,10))

  repeat
    cmd2 := RxSerByte
'  until cmd2 == "G" or cmd2 == "P" or
    if cmd2 == "G" or cmd2 == "P"
      quit
    if cmd2 == ESC or cmd2 == "A"
      return 0

  i~
  repeat
    ch := RxSerByte
    if i => 13                                          'FILENAMEBUFFERSIZE
      abort 'string("Filename too long")
    filename[i++] := ch
  while ch

  if cmd2 == "G"
    SendFileToPC
  else
    ReceiveFileFromPC

pri SendFileToPC | blksize, i, chksum, b, n
  if fsrw.pOpen( @filename, "R" ) <> 0
    text.str( @filename )
    abort 'string(" -- can't open file for reading")
  TxSerByte( "A" ) ' ack

  n~
  repeat
    blksize := fsrw.pRead( @buff2, 1024 )
    if blksize == -1
      blksize~
    TxSerByte( blksize.byte[0] )
    TxSerByte( blksize.byte[1] )
    i~
    chksum~
    repeat blksize
      b := buff2[i++]
      chksum += b
      TxSerByte( b )
    n += blksize
    if chksum & $ff <> RxSerByte
      abort 'string("Checksum error")
  while blksize == 1024
  fsrw.pClose
  text.dec( n )
  text.str( string(" bytes sent", 13) )

pri ReceiveFileFromPC | blksize, i, chksum, b, n
  if fsrw.pOpen( @filename, "W" ) <> 0
    text.str( @filename )
    abort 'string(" -- can't open file for writing")
  TxSerByte( "A" ) ' ack

  n~
  repeat
    blksize~
    blksize.byte[0] := RxSerByte
    blksize.byte[1] := RxSerByte
    if blksize > 1024
      abort string("Block size > 1024")
    i~
    chksum~
    repeat blksize
      b := RxSerByte
      buff2[i++] := b
      chksum += b
    n += blksize
    fsrw.pWrite( @buff2, blksize )
    TxSerByte( chksum )
  while blksize == 1024
  fsrw.pClose
  text.dec( n )
  text.str( string(" bytes received", 13) )

pri RxSerByte : b
' ser.rx with timeout
  b := key.getkey               'kbd.rxtime( TIMEOUT )
  if b == -1
    abort string("RxSerByte timed out")

pri TxSerByte( b )
' just for symmetry with RxSerByte
  text.out( b )

pri spin | i, j
  repeat i from 0 to 7
    j := byte[@inputbuffer+arg1+i]
    if j == 0 or j == "."
      quit
    else
      dirbuf[i] := j
  'force .bin extension
  strcopy(string(".BIN"),@dirbuf+i)
  if fsrw.popen(@dirbuf,"r") < 0
    text.str(string("Could not open file "))
    text.str(@dirbuf)
    text.str(string(".",cr))
  else
    fsrw.bootSDCard


pri doscolor(c)
  if c < 0 or c > 7
    text.str(string("Valid Colors are 0-7.",cr))
  else
    text.out($0c)
    text.out(48+c) 

pri clust2sd | aclust
  aclust := numfromarg(@inputbuffer+arg1)
  text.str(string("Cluster $"))
  text.hex(aclust,4)
  text.str(string(" at Sector $"))
  text.hex(fsrw.dos_clust2sd(aclust),6) 
  text.out(cr)

pri doschain | aclust, l, sn, sl, erc 
  aclust := numfromarg(@inputbuffer+arg1)
  optmore := checkoption(1,string("|MORE"))
  repeat
    l += 1
    if l>6
      lmore(0)
      l:=0
    text.hex(aclust,4)
    text.out(spc)
    if aclust<2 or aclust>$fff0
      if l <> 0
        lmore(0)
      quit
    sn := fsrw.dos_fat + (aclust >> 8)
    if sn<>sl
      erc := fsrw.readSDCard(sn, @secbuf, fsrw#SECTORSIZE)
      if erc < 0
        if l <> 0
          lmore(0)
        text.str(string("Could not read FAT sector $"))
        text.hex(sn,6)
        lmore(0)
        return
      sl := sn
    sn := (aclust & $ff) << 1
    aclust :=  byte[@secbuf+sn] + byte[@secbuf+sn+1]<<8        

PRI secdump | sdsec, erc, c, r, w, adx, v
  sdsec := numfromarg(@inputbuffer+arg1)
  optmore := checkoption(1,string("|MORE"))
  optmore := true                                       '<======always

  w := 8
  if checkoption(1,string("/W"))
    w := 16
  erc := fsrw.readSDCard(sdsec, @secbuf, fsrw#SECTORSIZE)
  if erc < 0
    text.str(string("Could not read SD sector $"))
    text.hex(sdsec,6)
    text.out(cr)
    return

  dosXS                                                 'dislay sector no

  text.str(string("SD Sector "))
  text.hex(sdsec,6)
  text.out(58)
  text.out(cr)
  adx := 0
  repeat while adx < fsrw#sectorsize
    text.hex(adx,4)
    text.str(string(":"))
    c := 0
    repeat while c < w
      v := adx + c
      text.out(spc)
      text.hex(secbuf[v],2)
      c++
    c := 0
    text.out(spc)
    repeat while c < w
      v := adx + c
      r := secbuf[v]
      c++
      if r =< 1 or (r => bksp and r =< cr)
        text.out(46)
      else
        text.out(r)
    adx += 8
    if not lmore(0)
      return
     
PRI dosdump | r, c, lc[15], w, adx, atend
  r := fsrw.popen(@inputbuffer+arg1,"r")
  if r < 0
    text.str(string("Could not open file: "))
    text.str(@inputbuffer+arg1)
    text.out(cr)
    return
  optmore := checkoption(1,string("|MORE"))
  w := 8
  if checkoption(1,string("/W"))
    w := 16
  adx := 0
  atend := false
  repeat until atend
    text.out(spc)
    text.hex(adx,4)
    text.str(string(":"))
    c := 0
    repeat while c < w  
      r := fsrw.pgetc
      if r < 0
        text.str(string("   "))
        atend := true
        lc[c] := -1
      else
        text.out(spc)
        text.hex(r,2)
        lc[c] := r
      c++
    c := 0
    text.out(spc)
    repeat while c < w
      r := lc[c]
      if r =< 1 or (r => bksp and r =< cr)
        text.out(46)
      else
        text.out(r)
      c++
    adx += 8
    if not lmore(0)
      return

'RR20090406
PRI dosdump2 | r, c, lc[15], w, adx, atend
  r := fsrw.popen(@inputbuffer+arg1,"r")
  if r < 0
    text.str(string("Could not open file: "))
    text.str(@inputbuffer+arg1)
    text.out(cr)
    return
  optmore := checkoption(1,string("|MORE"))
  w := 8
  if checkoption(1,string("/W"))
    w := 16
  adx := 0
  atend := false
  dispinfo                                              'display current cluster info
  text.out(cr)
  repeat until atend
    text.out(spc)
    text.hex(adx,4)
    text.str(string(":"))
    c := 0
    repeat while c < w  
      r := fsrw.pgetc
      if r < 0
        text.str(string("   "))
        atend := true
        lc[c] := -1
      else
        text.out(spc)
        text.hex(r,2)
        lc[c] := r
      c++
    c := 0
    text.out(spc)
    repeat while c < w
      r := lc[c]
      if r =< 1 or (r => bksp and r =< cr)
        text.out(46)
      else
        text.out(r)
      c++
    adx += 8
    dispinfo                                            'display current cluster info
    if not lmore(0)
      return

PRI dosx2 | i, j, s, z, ky, z0, n
' found at $80301..$90300 ($10000*128=8MB, $10000*512=32MB, 512B sectors, 16K clusters)
'look for sector
  z := 2 * 1024 / 512 * 1024 * 1024                     ' 2GB / 512
  z0 := $400 * 512
  n :=0
  text.out(cr)
  repeat s from z0 to z - 1
    fsrw.readSDCard(s, @buff, 512)                      's=the start byte within the file (not sector/cluster)
    if buff[10] == "*"
      text.hex(s/512,8)                                 ' s/512
      text.str(string(": "))
        repeat j from 0 to 7
          text.hex(buff[j],2)
          text.out(" ")
      text.out(cr)
      waitcnt(clkfreq/500+cnt)
      text.out("*")
      n++
      repeat while n < 100 'key.key <> "n"

PRI dosXS
  text.hex(ioControl[3],8)                              ' display sector no from sdspiFemto
  text.out(spc)
  text.hex(ioControl[0],8)
  text.out(spc)
  text.hex(ioControl[1],8)
  text.out(spc)
  text.hex(ioControl[2],8)
  text.out(cr)

PRI dispinfo | i
' display the other info sector# etc
  text.out(cr)
  fsrw.getpars(@x_fclust)               ' copy the 19 cluster info etc from fsrwFemto
  repeat i from 0 to 9
    text.hex(x_fclust[i],8)
    text.out(" ")
  text.out(cr)
  repeat i from 0 to 9                  ' only 9 not 10 !!
    text.hex(x_fclust[i+10],8)
    text.out(" ")

{
  text.hex(x_fclust, 6)
  text.out(" ")
  text.hex(x_filesize, 6)
  text.out(" ")
  text.hex(x_floc, 6)
  text.out(" ")
  text.hex(x_frem, 6)
  text.out(" ")
  text.hex(x_writelink, 6)
  text.out(" ")
  text.hex(x_dataregion, 6)
  text.out(" ")
  text.hex(x_clustershift, 6)
  text.out(" ")
  text.hex(x_totclusters, 6)
  text.out(" ")
  text.hex(x_sectorsperfat, 6)
  text.out(" ")
  text.hex(x_lastread, 6)
  text.out(cr)
} 


PRI dosdir | opthid, optc, r, i,  opt1, c, hidden, b, fs, x

  opt1 := 1
  optwc := false
  if word[@inputbuffer+arg1] == $2e2a  '*.
    optwc := true
    arg1 += 2
    opt1++
  opthid := checkoption(opt1,string("/H"))
  optc := checkoption(opt1,string("/C"))
  optmore := checkoption(opt1,string("|MORE"))
  optwide := checkoption(opt1,string("/W"))
  '
  text.out(cr)
  text.str(string("Directory of SD"))
  lmore(0)
  lmore(0)
  fsrw.opendir
  b := 0
  repeat
    'get next file    
    c := fsrw.nextfile(@dirbuf)
    if c == -1
      'we are finished
      if b > 0
        lmore(0)
      lmore(0)
      quit
      
    hidden := false
    
    if optwc
      'extract extension 
      repeat i from 0 to 2
        c := dirbuf[i+8]
        if c == spc
          fileext[i] := 0
        else
          fileext[i] := c
      'null terminate
      fileext[3] := 0
      'compare with pattern
      if not strcomp(@fileext,@inputbuffer+arg1)
        hidden := true

    if not opthid
      c := dirbuf[$B] 'attribute byte
      if c & 2        'hidden bit                
        hidden := true
   
    if not hidden

      if optwide
        'see if we need a CR
        c := 13
        if optc
          c += 5
        b += c
        if b > 39 
          if not lmore(0)
            return
          b := c
      else
        'display the file size
        'we can't use long[] unless the offset into the dirbuf is word aligned
        'this always works...
        fs := byte[@dirbuf+$1C] + byte[@dirbuf+$1D]<<8 + byte[@dirbuf+$1E]<<16 + byte[@dirbuf+$1F]<<24
        outdec(fs,10)
        text.out(spc)
        if opthid 
          c := dirbuf[$B] 'attribute byte
          if c & 2        'hidden bit
            text.out("H")
          else
            text.out(spc)                                                             
          text.out(spc)                                                             
           
      'always display the filename at fixed length
      r := 0
      repeat i from 0 to 7
        c := dirbuf[i]
        if c <> spc    
          text.out(c)
          r++
      if dirbuf[8] <> spc
        text.out(".")
        r++
        repeat i from 8 to 10
          c := dirbuf[i]
          if c <> spc    
            text.out(c)
            r++
      repeat while r < 13
        text.out(spc)
        r++

      if optc
        text.out("$")
        text.hex(byte[@dirbuf+$1B],2)
        text.hex(byte[@dirbuf+$1A],2)
        text.out(spc)

        x := (byte[@dirbuf+$1B] * 256) + (byte[@dirbuf+$1A])
        text.dec(x)
        text.out(spc)

        x := x * 512
        text.hex(x,8)
        text.out(spc)

        x := x * 16 * 1024
        text.hex(x,8)
        text.out(spc)

        text.out(cr)
        x := $80301 / $402
        text.hex(x,8)
        text.out(cr)
        
        

        
      if not optwide
        'displayed file size already, now cr/more
        if not lmore(0)
          return
     
pri dostype | r, c
  r := fsrw.popen(@inputbuffer+arg1,"r")
  if r < 0
    text.str(string("Could not open file: "))
    text.str(@inputbuffer+arg1)
    text.out(cr)
    return
  optmore := checkoption(1,string("|MORE"))
  r := 0
  repeat
    c := fsrw.pgetc
    if c < 0
      text.out(cr)
      fsrw.pclose
      return
    if c > 0
      if c == cr
        r := 0
        if not lmore(0)
          fsrw.pclose
          return
      elseif c>1 and (c < bksp or c > cr)
        text.out(c)
        r++
        if r == 40
          'wrapped to next line
          r := 0
          if not lmore(true)
            fsrw.pclose
            return

' functions used by commands
           
pri numfromarg(s) | nbase, c
  nbase := 16
  result := 0
  repeat 
    c := byte[s]
    s++
    if c == 0
      return
    elseif c == "$" '$=hex
      nbase := 16
    elseif c == "#"
      nbase := 10
    else
      if c => "a"
        c := c - constant("a"-10)
      elseif c => "A"
        c := c - constant("A"-10)
      elseif c => "0"
        c := c - "0"
      result := result * nbase + c
      
pri lmore(nocr) | c
  result := true
  if nocr == 0
    text.out(cr)
  c := key.key
  if c == ctlckey
    abort
  if optmore
    lcount++
    if lcount > screenrows - 2
      text.str(string("-more-"))
      c := key.getkey
'RR20090406      if c == esckey or c == ctlckey
      if c == esckey or c == ctlckey or c == esc or c == ctlc                   ' & esc & ctl-c
        text.str(string(" (aborted)",cr))
        return false
      repeat 6
        text.str(string(bksp,spc,bksp))
      lcount := 0
    
{ input collects up to size-1` characters of keyboard
  input, ending on CR or ESC and properly handling
  backspace. Returns cr, esc, or ctlz. }
     
pri input(buffer, size) | c, cursor
  'NB size is size of input buffer; user can enter
  '1 fewer characters than this due to null term
  byte[buffer] := 0
  c:=0
  cursor:=0
  text.out(uscore)
  repeat
    c := key.key
    if c <> 0

      if c <> esckey 
        esclock := 0
        
'RR20090406      if c == bkspkey
      if c == bkspkey or c == bksp
        if cursor > 0
        byte[buffer--] := 0
        cursor--
        text.str(string(bksp,spc,bksp,bksp,spc,bksp,uscore))
        c := 0

      elseif c == cr
        text.str(string(bksp,spc,bksp,cr))
        return c
         
'RR20090406      elseif c == esckey and not esclock
      elseif (c == esckey and not esclock) or c == esc
        text.str(string(bksp,spc,bksp))
        return esc
      
'RR20090406      elseif c == ctlzkey
      elseif c == ctlzkey or c == ctlz
        text.str(string(bksp,spc,bksp))
        return ctlz

      elseif c > 31 and c < 127
        if cursor < size
          text.out(bksp)
          text.out(c)
          text.out(uscore)
          byte[buffer++] := c
          cursor++
          byte[buffer] := 0
                  
{ parse locates up to 8 space delimited subfields in the
  buffer and returns pointers to them in inputparse[],
  setting inputfields to the number found. Found fields
  are zero terminated where they lay so they can be
  treated as separate ordinary strings. }

pri parse(buffer) | cursor, c
  'not all code checks to see if an option actually exists,
  'so default all results to point at the null terminator
  cursor := strsize(buffer)
  repeat c from 0 to 7
    inputparse[c] := cursor
  cursor := 0
  inputfields := 0
  'advance to next nonspace character
  'or find end of string
  repeat
    repeat
      c := byte[buffer][cursor++]    
      if c == 0
        return
      if c <> spc
        quit
    inputparse[inputfields] := cursor-1
    inputfields++
    if inputfields > 7
      return
    'advance to next space or find end
    repeat
      c := byte[buffer][cursor++]
      if c == 0
        return
      if c == spc
        byte[buffer][cursor-1] := 0
        quit    
        
' Check for an option.  We have to tell it the first argument
' in the split command line to check because this routine
' doesn't know how many fixed-position arguments might be
' required or found.
'
PRI CheckOption(FirstArg, Opt) | n
  n := 1
  repeat while n < inputfields
    if strcomp(@inputbuffer+inputparse[n],opt)
      result := true
    n++

' The text.dec method doesn't provide for fixed digits

PRI outdec(value,len) | j

  j := len
  byte [@numbuf+j] := 0
  j--
  repeat
    byte[@numbuf+j] := value // 10 + "0"
    j--
    value /= 10
    if value == 0 or j == 0
      quit
  repeat
    if j < 0
      quit
    byte [@numbuf+j] := spc
    j--
  text.str(@numbuf)
       
PRI strcopy(src,dest) | c
  c := -1
  repeat while c <> 0
    c := byte[src]
    byte[dest] := c
    src++
    dest++

PRI ucase(astr) | c
  repeat
    c := byte[astr]
    if c == 0
      quit
    if c => "a" and c =< "z"
        byte[astr] := c - constant("a" - "A")
    astr++

dat
''
'' This DAT block can be omitted if using the aigeneric driver
'' or the default palette for tv_text.

alt_palette     byte    $07,   $02    '0    white / black (background)
                byte    $07,   $02    '1    white / black
                byte    $6B,   $02    '2    green / black
                byte    $0A,   $02    '3     blue / black
                byte    $02,   $07    '4    black / white
                byte    $BB,   $07    '5      red / white
                byte    $07,   $BB    '6    white / red
                byte    $07,   $0A    '7    white / blue

{{

                                                   TERMS OF USE: MIT License                                                                                                              

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation     
files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    
modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:                                                                   
                                                                                                                              
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
                                                                                                                              
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         
}}
